% ENERGY    computes the error and the jacobian of P function (maps points
% on the 2D template on the deformed paper). The error computed is the
% union of:
%       - reprojection error: the error obtained by projecting the points
%         belonging to the template on the final image, exploiting P function
%         and the camera calibration matrix
%       - smoothness error: the second derivative on the deformed edges, it
%         has to be minimized to obtain a smooth final image
%       - lower edge error: to maintain the length on the lower deformed
%         edge
%       - upper edge error: to maintain the length on the upper deformed
%         edge
%       - width error: to maintain the heigth of the image
%
%   [e, j]=energy(S, dataStruct) computes e (error) and j (jacobian)
%   given:
%   x is a 3*M column vector, containing the coordinates of the vertices of the
%       triangles (even points must be equi-spaced). The vector must be in
%       this form: x=[x1 x2 x3 ... xm y1 y2 ... ym z1 z2 ... zm]'. M must
%       be >=4 to properly approximate the rectangular paper
%   dataStruct contains the following fields:
%       S camera calibration matrix 
%       tPoints is a Nx3 matrix, each row contains the homogeneous coordinates
%           of the points on the original template
%       iPoints is a Nx3 matrix, each row contains the homogeneous coordinates
%           of the points on the final image
%       W is the straight edge length
%       H is the triangles height
%   Pamela Gotti - pamela.gotti@gmail.com


function [e, j]=energy(x, datastruct)

if size(x,1)<12
    error 'Input argument x needs at least 12 values (3 coordinates - x,y,z, at least 4 variables)'
end
if mod(size(x, 1), 3) ~= 0
    error 'Input argument x size must be multiple of 3 (3 coordinates - x,y,x - per variable)'
end

%% setting global constant and function variables
global R_CONST; %reprojection error weight
global S_CONST; %smoothness error weight
global M1_CONST; %length preservation on lower edge weight
global M2_CONST; %length preservation on upper length weight
global M3_CONST; %width preservation weight
%this three variables are posed global to allow the solution projection
global barCoord;
global verTri;
global e1;

% aPoints is a Nx4 matrix containing the homogeneous coordinates of the
% vertices of the 3D triangles in the form
% aPoints=[x1 y1 z1 1;
%          x2 y2 z2 1;
%               ...
%          xn yn zn 1];
aPoints=x;
aPoints=reshape(aPoints, size(x,1)/3, 3);
aPoints=[aPoints ones(size(aPoints, 1), 1)];
S=datastruct.S;         % camera calibration matrix
tPoints=datastruct.tPoints;   % points on the template
iPoints=datastruct.iPoints;   % points on the image with the deformed paper
W=datastruct.W;                 % template width
H=datastruct.H;                 % template height
% calculation of the projection of the aPoints on the 2D space
p2d=zeros(size(aPoints, 1), 3);
p2d(1,:)=[0, H, 1];
p2d(2,:)=[0 0 1];
for i=1:size(aPoints, 1)-2
    % calculation of the distance between points i and i+2
    d=sqrt((aPoints(i,1)-aPoints(i+2,1))^2+(aPoints(i,2)-aPoints(i+2,2))^2+(aPoints(i,3)-aPoints(i+2,3))^2);
    % projection of the 3d point on the 2d space
    p2d(i+2,:)=p2d(i,:)+[d, 0, 0];
end
% computation of barCoord and verTri
%   barCoord is a Nx3 matrix, each line contains the barycentric
%       coordinates for the template points: u=a1*v1+a2*v2+a3*v3, u is the
%       point, v1, v2, and v3 the coordinates of the triangle the point
%       belongs to, a1, a2, a4 the barycentric coordinates
%   verTri is a Nx3 matrix, each line contains the vertices number of the triangle
%      the point belong to, in ascending order. The vertices are enumerated considering them
%      as a broken line starting from the top left angle of the template:
%       1----3---5-------7     
%       |   /\   | \     |
%       |  /  \  |  \    |
%       | /    \ |  * \  |
%       |/      \|     \ |
%       2--------4-------6         
%       in example, for point * verTri will contain the row [4 5 6]        
barCoord=zeros(size(tPoints,1), 3);
verTri=zeros(size(tPoints, 1), 3);
A=zeros(3, 3);
for i=1:size(tPoints, 1)
    for k=1:size(p2d, 1)-2
        A(1:2,1:3)=p2d(k:k+2, 1:2)';
        A(3,:)=ones(1, 3);
        b=A\tPoints(i,:)';   %barycentric coordinates of i-th point respect to k-th triangle; if one of them is <0 the point doesn't belong to the triangle
        if sum(b<-eps)==0
            barCoord(i,:)=b';
            verTri(i,:)=k:k+2;
            break
        end
    end
end
clear A

if sum(verTri==0)==0    %if the condition is satisfied, all template's points belong to one triangle
%% ed computation: reprojection error
    M=aPoints(verTri',:);
    B=repmat(reshape(barCoord',numel(barCoord),1), 1, 4);
    T=(B.*M)';
    T=reshape(T', 3, 4*size(tPoints,1));
    T=sum(T);
    T=reshape(T, size(tPoints,1), 4);
    u=(S*T')';
    ed=iPoints-u./repmat(u(:,3), 1, 3);
    ed=reshape(ed, numel(ed), 1);
    ed=ed(1: end-size(tPoints,1));


%% jacobian computation: reprojection error
    % terms about reprojection error are structured in this way:
    % 2xN rows, N is the number of template points:
    % ux-(s11*(a(n)*x(n)+a(n+1)*x(n+1)+a(n+2)*x(n+2))+s12*(a(n)*y(n)+a(n+1)*y(n+1)+a(n+2)*y(n+3))+s13*(a(n)*z(n)+a(n+1)*z(n+1)+a(n+2)*z(n+2))+s14)/(s31*(a(n)*x(n)+a(n+1)*x(n+1)+a(n+2)*x(n+2))+s32*(a(n)*y(n)+a(n+1)*y(n+1)+a(n+2)*y(n+2))+s33*(a(n)*z(n)+a(n+1)*z(n+1)+a(n+2)*z(n+2))+s34)
    % uy-(s21*(a(n)*x(n)+a(n+1)*x(n+1)+a(n+2)*x(n+2))+s22*(a(n)*y(n)+a(n+1)*y(n+1)+a(n+2)*y(n+3))+s23*(a(n)*z(n)+a(n+1)*z(n+1)+a(n+2)*z(n+2))+s24)/(s31*(a(n)*x(n)+a(n+1)*x(n+1)+a(n+2)*x(n+2))+s32*(a(n)*y(n)+a(n+1)*y(n+1)+a(n+2)*y(n+2))+s33*(a(n)*z(n)+a(n+1)*z(n+1)+a(n+2)*z(n+2))+s34)
    % n, n+1, n+2 are defined by the verTri variable, [ux uy 1] are the
    % coordinates of the template point
    % let D=(s31*(a(n)*x(n)+a(n+1)*x(n+1)+a(n+2)*x(n+2))+s32*(a(n)*y(n)+a(n+1)*y(n+1)+a(n+2)*y(n+2))+s33*(a(n)*z(n)+a(n+1)*z(n+1)+a(n+2)*z(n+2))+s34)
    % and N1=(s11*(a(n)*x(n)+a(n+1)*x(n+1)+a(n+2)*x(n+2))+s12*(a(n)*y(n)+a(n+1)*y(n+1)+a(n+2)*y(n+3))+s13*(a(n)*z(n)+a(n+1)*z(n+1)+a(n+2)*z(n+2))+s14)
    % and N1=(s21*(a(n)*x(n)+a(n+1)*x(n+1)+a(n+2)*x(n+2))+s22*(a(n)*y(n)+a(n+1)*y(n+1)+a(n+2)*y(n+3))+s23*(a(n)*z(n)+a(n+1)*z(n+1)+a(n+2)*z(n+2))+s24)
    % since the barycentric coordinates depend from the vertices coordinates, they are considered too in the computation of the derivative:
    % a1=(x(n+2)*y-x*y(n+2)-x(n+1)*y+y(n+1)*x+x(n+1)*y(n+2)-y(n+1)*x(n+2))/(-x(n+1)*y(n)+x(n+1)*y(n+2)-x(n)*y(n+2)+y(n)*x(n+2)-y(n+1)*x(n+2)+y(n+1)*x(n))
    % a2=(x(n)*y-x(n)*y(n+2)-x(n+2)*y(n)+y*x(n+2)-x*y(n+2)+y(n)*x)/(-x(n+1)*y(n)+x(n+1)*y(n+2)-x(n)*y(n+2)+y(n)*x(n+2)-y(n+1)*x(n+2)+y(n+1)*x(n))
    % a3=(-x(n+1)*y(n)+x(n)*y(n+1)+x(n+1)*y-y*x(n)+x*y(n)-y(n+1)*x)/(-x(n+1)*y(n)+x(n+1)*y(n+2)-x(n)*y(n+2)+y(n)*x(n+2)-y(n+1)*x(n+2)+y(n+1)*x(n))
    Jed=zeros(size(ed,1), size(aPoints,1)*3);
    D=repmat(u(:,3), 1, 3*size(aPoints,1));
    N1=repmat(u(:,1), 1, 3*size(aPoints,1));
    N2=repmat(u(:,2), 1, 3*size(aPoints,1));
    a=verTri(:,1);
    Sx=[-aPoints(a, 1) aPoints(a,1) -aPoints(a+1,1) aPoints(a+1, 1) aPoints(a+2, 1) -aPoints(a+2, 1)];
    Dx=[aPoints(a+2, 2) aPoints(a+1,2) aPoints(a,2) aPoints(a+2, 2) aPoints(a, 2) aPoints(a+1, 2)];
    DA=(sum((Sx.*Dx)'))';
    nA1=aPoints(a+2,1).*tPoints(:,2)-tPoints(:,1).*aPoints(a+2, 2)-aPoints(a+1, 1).*tPoints(:,2)+aPoints(a+1,2).*tPoints(:,1)+aPoints(a+1, 1).*aPoints(a+2, 2)-aPoints(a+1, 2).*aPoints(a+2, 1);
    nA2=aPoints(a,1).*tPoints(:,2)-aPoints(a,1).*aPoints(a+2, 2)+aPoints(a, 2).*aPoints(a+2,1)-aPoints(a+2,1).*tPoints(:,2)+aPoints(a+2, 2).*tPoints(:, 1)-aPoints(a, 2).*tPoints(:, 1);
    nA3=-aPoints(a+1,1).*aPoints(a,2)+aPoints(a+1,2).*aPoints(a, 1)+aPoints(a+1, 1).*tPoints(:,2)-aPoints(a,1).*tPoints(:,2)+aPoints(a, 2).*tPoints(:, 1)-aPoints(a+1, 2).*tPoints(:, 1);
    % derivatives of the projection on 2d space respect to x coordinates
    dX=zeros(size(tPoints,1), 3*size(aPoints, 1));
    % first line of dX is zero, since X1 is constant; so is the second line
    for i=3:max(max(verTri))
        dX(i,:)=dX(i-2,:);
        DX=sqrt((aPoints(i-2, 1)-aPoints(i, 1))^2+(aPoints(i-2, 2)-aPoints(i, 2))^2+(aPoints(i-2, 3)-aPoints(i, 3))^2);
        dX(i, i-2:i)=dX(i, i-2:i)+[-(aPoints(i-2, 1)-aPoints(i, 1))/DX 0 (aPoints(i-2,1)-aPoints(i,1))/DX];
        dX(i,size(aPoints,1)+i-2:size(aPoints,1)+i-2+3-1)=dX(i,size(aPoints,1)+i-2:size(aPoints,1)+i-2+3-1)+[-(aPoints(i-2, 2)-aPoints(i, 2))/DX 0 (aPoints(i-2,2)-aPoints(i,2))/DX];
        dX(i,2*size(aPoints,1)+i-2:size(aPoints,1)*2+3+i-2-1)=dX(i,2*size(aPoints,1)+i-2:size(aPoints,1)*2+3+i-2-1)+[-(aPoints(i-2, 3)-aPoints(i, 3))/DX 0 (aPoints(i-2,3)-aPoints(i,3))/DX];
    end
    % derivatives of the barycentric coordinates
    dA1=(dX(verTri(:,2), :).*repmat(p2d(verTri(:,3), 2),1,size(dX, 2))-(dX(verTri(:,2), :).*repmat(tPoints(:,2),1,size(dX, 2)))+(dX(verTri(:,3), :).*repmat(tPoints(:,2),1,size(dX, 2)))-(dX(verTri(:,3), :).*repmat(p2d(verTri(:,2),2),1,size(dX, 2)))).*repmat(DA, 1, size(aPoints,1)*3)-repmat(nA1, 1, size(aPoints,1)*3).*(dX(verTri(:,1), :).*repmat(p2d(verTri(:,2), 2),1,size(dX, 2))-(dX(verTri(:,1), :).*repmat(p2d(verTri(:,3),2),1,size(dX, 2)))+(dX(verTri(:,2), :).*repmat(p2d(verTri(:,3),2),1,size(dX, 2)))-(dX(verTri(:,2), :).*repmat(p2d(verTri(:,1),2),1,size(dX, 2)))+repmat(p2d(verTri(:,1),2),1,size(dX, 2)).*dX(verTri(:,3), :)-(repmat(p2d(verTri(:,2), 2),1,size(dX, 2)).*dX(verTri(:,3),:)));
    dA2=(dX(verTri(:,1), :).*repmat(tPoints(:,2),1,size(dX, 2))-(dX(verTri(:,1), :).*repmat(p2d(verTri(:,3),2),1,size(dX, 2)))+(dX(verTri(:,3), :).*repmat(p2d(verTri(:,1),2),1,size(dX, 2)))-(dX(verTri(:,3), :).*repmat(tPoints(:,2),1,size(dX, 2)))).*repmat(DA, 1, size(aPoints,1)*3)-repmat(nA2, 1, size(aPoints,1)*3).*(dX(verTri(:,1), :).*repmat(p2d(verTri(:,2), 2),1,size(dX, 2))-(dX(verTri(:,1), :).*repmat(p2d(verTri(:,3),2),1,size(dX, 2)))+(dX(verTri(:,2), :).*repmat(p2d(verTri(:,3),2),1,size(dX, 2)))-(dX(verTri(:,2), :).*repmat(p2d(verTri(:,1),2),1,size(dX, 2)))+repmat(p2d(verTri(:,1),2),1,size(dX, 2)).*dX(verTri(:,3), :)-(repmat(p2d(verTri(:,2), 2),1,size(dX, 2)).*dX(verTri(:,3),:)));
    dA3=(dX(verTri(:,1), :).*repmat(p2d(verTri(:,2), 2),1,size(dX, 2))-(dX(verTri(:,1), :).*repmat(tPoints(:,2),1,size(dX, 2)))+(dX(verTri(:,2), :).*repmat(tPoints(:,2),1,size(dX, 2)))-(dX(verTri(:,2), :).*repmat(p2d(verTri(:,1),2),1,size(dX, 2)))).*repmat(DA, 1, size(aPoints,1)*3)-repmat(nA3, 1, size(aPoints,1)*3).*(dX(verTri(:,1), :).*repmat(p2d(verTri(:,2), 2),1,size(dX, 2))-(dX(verTri(:,1), :).*repmat(p2d(verTri(:,3),2),1,size(dX, 2)))+(dX(verTri(:,2), :).*repmat(p2d(verTri(:,3),2),1,size(dX, 2)))-(dX(verTri(:,2), :).*repmat(p2d(verTri(:,1),2),1,size(dX, 2)))+repmat(p2d(verTri(:,1),2),1,size(dX, 2)).*dX(verTri(:,3), :)-(repmat(p2d(verTri(:,2), 2),1,size(dX, 2)).*dX(verTri(:,3),:)));
    %preparation of the barycentric coordinates
    indici=repmat((0:size(ed,1)/2-1)', 1, 3);
    indici=indici'*size(aPoints, 1)+verTri';
    bAr=zeros(size(aPoints,1), size(ed,1)/2);
    bAr(reshape(indici, numel(indici), 1))=reshape(barCoord',numel(barCoord),1);
    bAr=bAr';
    bAr=[bAr bAr bAr];
    Esse=[S(1,1)*ones(size(verTri,1), size(aPoints,1)) S(1,2)*ones(size(verTri,1), size(aPoints,1)) S(1,3)*ones(size(verTri,1), size(aPoints,1))];
    dN1=-(Esse.*(dA1.*repmat(aPoints(verTri(:,1), 1), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 1), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 1), 1,size(dA1, 2))+bAr)+S(1,2)*(dA1.*repmat(aPoints(verTri(:,1), 2), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 2), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 2), 1,size(dA1, 2)))+S(1,3)*(dA1.*repmat(aPoints(verTri(:,1), 3), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 3), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 3), 1,size(dA1, 2))));
    Esse=[S(2,1)*ones(size(verTri,1), size(aPoints,1)) S(2,2)*ones(size(verTri,1), size(aPoints,1)) S(2,3)*ones(size(verTri,1), size(aPoints,1))];
    dN2=-(Esse.*(dA1.*repmat(aPoints(verTri(:,1), 1), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 1), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 1), 1,size(dA1, 2))+bAr)+S(2,2)*(dA1.*repmat(aPoints(verTri(:,1), 2), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 2), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 2), 1,size(dA1, 2)))+S(2,3)*(dA1.*repmat(aPoints(verTri(:,1), 3), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 3), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 3), 1,size(dA1, 2))));
    Esse=[S(3,1)*ones(size(verTri,1), size(aPoints,1)) S(3,2)*ones(size(verTri,1), size(aPoints,1)) S(3,3)*ones(size(verTri,1), size(aPoints,1))];
    dD=-(Esse.*(dA1.*repmat(aPoints(verTri(:,1), 1), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 1), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 1), 1,size(dA1, 2))+bAr)+S(3,2)*(dA1.*repmat(aPoints(verTri(:,1), 2), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 2), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 2), 1,size(dA1, 2)))+S(3,3)*(dA1.*repmat(aPoints(verTri(:,1), 3), 1,size(dA1, 2))+dA2.*repmat(aPoints(verTri(:,2), 3), 1,size(dA1, 2))+dA3.*repmat(aPoints(verTri(:,3), 3), 1,size(dA1, 2))));
    
    Jed1=(dN1.*D-dD.*N1)./(D.^2);
    Jed(1:2:end,:)=Jed1;
    Jed1=(dN2.*D-dD.*N2)./(D.^2);
    Jed(2:2:end,:)=Jed1;
    clear M B T u P1 P D N1 N2 Jed1 Esse indici DA dA nA1 nA2 nA3 dA1 dA2 dA3 Sx Dx

%% es computation: smoothness factor
    if size(aPoints,1)>4    %possible only if there are at least 5 points (the second derivative is computed on 3 adjacent vertices on the same edge)
        M1=aPoints(1:end-4,1:3);
        M2=aPoints(3:end-2,1:3);
        M3=aPoints(5:end, 1:3);
        es=M1-2*M2+M3;
        es=reshape(es, numel(es), 1);
        %% jacobian computation: smoothness factor
        % terms about smoothness factor are structured this way:
        % for x coordinate x(n)-2*x(n+2)+x(n+4), so the derivative is like [... 0 0 1 0 -2
        % 0 1 0 0 0 ...]
        % the same for y and z
        % there are 3*(N-4) error terms, N is the number of variable
        D1=[diag(ones((size(aPoints,1)-4),1)) zeros((size(aPoints,1)-4), 4)];
        D2=diag(-2*ones((size(aPoints,1)-4),1), 2);
        D2=[D2(1:size(aPoints,1)-4, :), zeros((size(aPoints,1)-4), 2)];
        D3=diag(ones(size(aPoints,1)-4,1), 4);
        D3=D3(1:size(aPoints,1)-4, :);
        Jes1=D1+D2+D3;
        Jes=[Jes1 zeros(size(aPoints)-4, size(aPoints)*2); zeros(size(aPoints)-4, size(aPoints)) Jes1 zeros(size(aPoints)-4, size(aPoints)); zeros(size(aPoints)-4, size(aPoints)*2) Jes1];
        clear M1 M2 M3 D1 D2 D3 Jes1
    else        % if no second derivative calculation is possible, there is no error
        es=0;
        Jes=zeros(1, size(aPoints, 1)*3);
    end


%% em computation: metric error1
    % length preservation on lower edge (odd points on the broken line); the points on this edge are equi-spaced
    %distance between two points
    N=(W/(floor(size(aPoints,1)/2)-1))^2;
    M1=aPoints(2:2:end-2,1:3);
    M2=aPoints(4:2:end, 1:3);
    em1=dot((M1-M2)',(M1-M2)');
    em1=em1'-N*ones(size(em1,1),1);

%% jacobian computation: metric error 1
    % only odd terms are non zero
    % error terms are like (x(n+2)-x(n))^2+(y(n+2)-y(n))^2+(z(n+2)-z(n))^2-N^2
    D=2*[-(M2-M1) (M2-M1)];
    Jm1=zeros(size(em1,1), 3*size(aPoints,1));
    indici=size(em1,1)+1:2*size(em1,1)+1:size(em1,1)*(size(aPoints,1)-1);
    somma=size(aPoints,1)*size(em1,1);
    s=[indici somma*ones(1, size(indici,2))+indici 2*somma*ones(1, size(indici,2))+indici];
    Jm1(s)=D(1:size(em1,1)*3);
    s=s+2*size(em1,1)*ones(1,size(s,2));
    Jm1(s)=D(3*size(em1,1)+1:end);
    clear M1 M2 D indici somma s

%% metric error 2 computation
    %length preservation on upper edge (even points on the broken line); points
    %on this edge are not equi-spaced
    M1=aPoints(1:2:end-2,1:3);
    M2=aPoints(3:2:end, 1:3);
    A=M1-M2;
    B=M2-M1;
    em=dot(A',A');
    em=em.^(1/2);
    em2=sum(em)-W;

%% jacobian computation: metric error 2
    % only even terms are non zero
    % error terms are like
    % sum(sqrt((x(n+2)-x(n))^2+(y(n+2)-y(n))^2+(z(n+2)-z(n))^2))
    A1=zeros(1, size(aPoints, 1)*3);
    B1=zeros(1, size(aPoints, 1)*3);
    indici=[1:2:size(aPoints, 1)-2 (size(aPoints,1))*ones(1, ceil(size(aPoints, 1)/2)-1)+(1:2:(size(aPoints, 1)-2)) (2*size(aPoints,1))*ones(1, ceil(size(aPoints, 1)/2)-1)+(1:2:(size(aPoints, 1)-2))];
    A1(indici)=-A(1:end)./repmat(em, 1, 3);
    indici=2*ones(1,size(indici))+indici;
    B1(indici)=-B(1:end)./repmat(em,1,3);
    Jm2=A1+B1;
    clear M1 M2 em A B A1 B1

%% error computation: metric error 3
    %width preservation
    em3=zeros(size(aPoints,1)-2, 1);
    for i=3:(size(aPoints,1))
        %triangle height
        h=norm(cross((aPoints(i-2,1:3)-aPoints(i,1:3)),((aPoints(i,1:3)-aPoints(i-1,1:3)))))/norm((aPoints(i-2,1:3)-aPoints(i,1:3)));
        % error computed as difference between the real height and the
        % approximated one (squared to avoid negative numbers)
        err=(h^2-H^2);
        em3(i-2)=err;
    end
    clear err

%% jacobian computation: metric error 3
    % error terms are like 
    % (|(P(i-2)-P(i))x(P(i)-P(i-1))|/|P(i-2)-P(i)|)^2-H^2 = ( ( ((y0-y2)*(z2-z1) - (z0-z2)*(y2-y1))^2 + ((z0-z2)*(x2-x1) - (x0-x2)*(z2-z1))^2 + ((x0-x2)*(y2-y1) - (y0-y2)*(z2-z1))^2 ) / ( (x0-x2)^2 + (y0-y2)^2 + (z0-z2)^2 ))-H^2;
    Jm3=zeros(size(em3,1), size(aPoints,1)*3);
    for i=3:size(aPoints,1)
        T=cross((aPoints(i-2,1:3)-aPoints(i,1:3)),((aPoints(i,1:3)-aPoints(i-1,1:3))));
        P=[(aPoints(i,1:3)-aPoints(i-2,1:3))'; (aPoints(i,1:3)-aPoints(i-1,1:3))'; (aPoints(i-1,1:3)-aPoints(i-2,1:3))'];
        N=(norm(T))^2;
        D=(norm((aPoints(i-2,1:3)-aPoints(i,1:3))))^2;
        dN=2*(reshape(repmat([T(2) T(1) T(1)], 3, 1), 9, 1).*[-P(6) P(3) -P(9) P(6) -P(3) P(9) -P(5) P(2) -P(8)]' + reshape(repmat([T(3) T(3) T(2)], 3, 1), 9, 1).*[P(5) -P(2) P(8) -P(4) P(1) -P(7) P(4) -P(3) P(7)]');
        dD=2*[-P(1) 0 P(1) -P(2) 0 P(2) -P(3) 0 P(3)];
        dF=(dN*D-N*dD')./(D^2);
        indici=[i-2:i size(aPoints,1)*ones(1, 3)+((i-2):i) 2*size(aPoints,1)*ones(1, 3)+((i-2):i)];
        Jm3((i-2), indici)=dF';
    end


    %error array
    e=[R_CONST*ed;S_CONST*es;M1_CONST*em1;M2_CONST*em2;M3_CONST*em3];
    e1=e;
    %jacobian array
    j=[Jed; Jes; Jm1; Jm2; Jm3];
else    %in the case come point doesn't belong to any triangle, high error value is returned
    n=size(tPoints, 1)*2+(size(aPoints,1)>4*size(aPoints,1)-4)+(size(aPoints,1)<=4*1)+floor((size(aPoints,1)-2)/2)+1+size(aPoints, 1)-2;
    e=1e10*ones(n, 1);
    j=zeros(n, size(aPoints, 1)*3);
end